from functools import partial
from typing import Any


def maybe_bind_self(func: Any, target_self: Any):
    if "self" in func.__code__.co_varnames:
        return partial(func, target_self)
    return func



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Awaitable, Protocol

from keyweave.commanding import FuncHotkeyHandler
from keyweave.hotkey import HotkeyEvent
from keyweave.shorthand import SimpleCoroutine
from keyweave.util.event_loop import norm_maybe_async

if TYPE_CHECKING:
    from keyweave.layout._layout_class import LayoutClass


@dataclass(init=False)
class HotkeyInterceptionEvent(HotkeyEvent):
    """
    Represents an intercepted `HotkeyEvent`. Used in the layout hotkey interception API.
    """

    def __str__(self):
        return super().__str__() + " (intercepted)"

    _handled: bool = False

    def __init__(self, event: HotkeyEvent, handler: FuncHotkeyHandler):
        HotkeyEvent.__init__(self, event.binding, event.event)
        self._handler = handler

    def next(self):
        self._handled = True
        result = self._handler(self)
        return result

    def end(self):
        self._handled = True

    @property
    def handled(self):
        return self._handled


class HotkeyInterceptor(Protocol):
    """
    A function that can intercept a hotkey event.
    """

    def __call__(
        self, action: HotkeyInterceptionEvent
    ) -> None | SimpleCoroutine[None]: ...


class FuncHotkeyInterceptor(Protocol):
    """
    A function that can intercept a hotkey event.
    """

    def __call__(
        self, action: HotkeyInterceptionEvent, /
    ) -> None | SimpleCoroutine[None]: ...


class MethodHotkeyInterceptor(Protocol):
    """
    A method that can intercept a hotkey event.
    """

    def __call__(
        self, instance: Any, action: HotkeyInterceptionEvent, /
    ) -> None | SimpleCoroutine[None]: ...


type AnyHotkeyInterceptor = (FuncHotkeyInterceptor | MethodHotkeyInterceptor)



import threading
from typing import Any
from keyweave.bindings import KeyBindingCollection
from keyweave.hotkey import InputEvent
from keyweave.key_types import Key, KeyInputState, KeySet
from keyboard import KeyboardEvent, hook_key, is_pressed, unhook

from keyweave._util.logging import keyweaveLogger

from keyweave.scheduling import Scheduler
from keyweave.util.input import get_keyboard_hook_id, is_key_pressed

keyweaveHookLogger = keyweaveLogger


class KeyHook:
    _internal_handler: Any
    _lock = threading.Lock()

    @property
    def _logger(self):
        return keyweaveHookLogger

    def __init__(
        self,
        key: Key,
        collection: KeyBindingCollection,
        scheduler: Scheduler,
    ):
        self.trigger = key
        self._collection = collection
        self._scheduler = scheduler
        self.manufactured_handler = self._get_handler()

    def __enter__(self):
        self._logger.info("Installing hook for %s", self.trigger)
        self._internal_handler = hook_key(
            get_keyboard_hook_id(self.trigger.id),
            self.manufactured_handler,
            suppress=True,
        )
        return self

    def __exit__(self):
        unhook(self._internal_handler)
        return self

    def _get_handler(self):
        def get_best_binding(event: KeyboardEvent):
            only_matching = [
                binding
                for binding in self._collection
                if binding.hotkey.trigger.key.is_numpad == event.is_keypad
                and _is_key_set_active(binding.hotkey.modifiers)
                and event.event_type == binding.hotkey.trigger.state
            ]

            by_specificity = sorted(
                only_matching,
                key=lambda binding: binding.hotkey.specificity,
                reverse=True,
            )
            return by_specificity[0] if by_specificity else None

        def handler(info: KeyboardEvent):
            binding = get_best_binding(info)
            event_state = _state_from_event(info)
            if not binding:
                return True
            self._logger.debug(f"RECEIVED {event_state}; MATCHED {binding}")

            def binding_invoker():
                return binding(InputEvent(info.time))

            self._scheduler(binding_invoker)
            return False

        return handler


def _is_state_active(state: KeyInputState):
    match state.state:
        case "down":
            return is_key_pressed(state.key.id)
        case "up":
            return not is_key_pressed(state.key.id)
        case _:
            raise ValueError(f"Unknown key state: {state.state}")


def _is_key_set_active(key_set: KeySet):
    return all(_is_state_active(key_state) for key_state in key_set)


def _state_from_event(event: KeyboardEvent) -> KeyInputState:
    return KeyInputState(
        key=Key(event.name or ""), state=event.event_type or "down"
    )



from keyweave.key_types import Key


num_1 = Key("num:1")
num_2 = Key("num:2")
num_3 = Key("num:3")
num_4 = Key("num:4")
num_5 = Key("num:5")
num_6 = Key("num:6")
num_7 = Key("num:7")
num_8 = Key("num:8")
num_9 = Key("num:9")
num_0 = Key("num:0")
num_enter = Key("num:enter")
num_plus = Key("num:+")
num_minus = Key("num:-")
num_asterisk = Key("num:*")
num_slash = Key("num:/")
num_dot = Key("num:.")

d_0 = Key("0")
d_1 = Key("1")
d_2 = Key("2")
d_3 = Key("3")
d_4 = Key("4")
d_5 = Key("5")
d_6 = Key("6")
d_7 = Key("7")
d_8 = Key("8")
d_9 = Key("9")
a = Key("a")
b = Key("b")
c = Key("c")
e = Key("e")
d = Key("d")
f = Key("f")
g = Key("g")
h = Key("h")
i = Key("i")
j = Key("j")
k = Key("k")
l = Key("l")
m = Key("m")
n = Key("n")
o = Key("o")
p = Key("p")
q = Key("q")
r = Key("r")
s = Key("s")
t = Key("t")
u = Key("u")
v = Key("v")
w = Key("w")
x = Key("x")
y = Key("y")
z = Key("z")

enter = Key("enter")
space = Key("space")
backspace = Key("backspace")
tab = Key("tab")
shift = Key("shift")
ctrl = Key("ctrl")
alt = Key("alt")
grave = Key("`")
minus = Key("-")
equal = Key("=")
plus = Key("+")
slash = Key("/")
backslash = Key("\\")
comma = Key(",")
dot = Key(".")
semicolon = Key(";")
quote = Key("'")
bracket_left = Key("[")
bracket_right = Key("]")

f1 = Key("f1")
f2 = Key("f2")
f3 = Key("f3")
f4 = Key("f4")
f5 = Key("f5")
f6 = Key("f6")
f7 = Key("f7")
f8 = Key("f8")
f9 = Key("f9")
f10 = Key("f10")
f11 = Key("f11")
f12 = Key("f12")

insert = Key("insert")
delete = Key("delete")
home = Key("home")
end = Key("end")
page_up = Key("page up")
page_down = Key("page down")

capslock = Key("capslock")
mouse_1 = Key("mouse:1")
mouse_2 = Key("mouse:2")
mouse_3 = Key("mouse:3")



from dataclasses import dataclass, field
from functools import partial
import inspect
from types import MethodType
from typing import (
    Any,
    Awaitable,
    Callable,
    Literal,
    NotRequired,
    Protocol,
    TYPE_CHECKING,
    TypedDict,
    Unpack,
)


from keyweave.shorthand import SimpleCoroutine
from keyweave.util.dict import default_dicts, merge_dicts
from keyweave.util.reflect import (
    clean_method_name,
    get_self_dict_restricted,
    is_unbound_method,
)


if TYPE_CHECKING:
    from keyweave.hotkey import Hotkey, HotkeyEvent
    from keyweave.layout._layout_class import LayoutClass
    from keyweave.interception import (
        FuncHotkeyInterceptor,
        AnyHotkeyInterceptor,
        HotkeyInterceptor,
        MethodHotkeyInterceptor,
    )


class CommandMetaProperties(TypedDict):
    label: NotRequired[str | None]
    description: NotRequired[str | None]
    emoji: NotRequired[str | None]


@dataclass(kw_only=True)
class CommandMeta:
    """
    Metadata about a command. Used for debugging and user feedback.
    """

    label: str | None = field(default=None)
    description: str | None = field(default=None)
    emoji: str | None = field(default=None)
    metadata: Any | None = field(default=None)

    def __str__(self):
        text = list[str]()
        if self.emoji:
            text.append(self.emoji)
        if self.label:
            text.append(self.label)
        return " ".join(text) if text else "<unnamed command>"

    def with_defaults(self, **kwargs: Unpack[CommandMetaProperties]):
        attrs = default_dicts(
            get_self_dict_restricted(self, CommandMeta), kwargs
        )
        return CommandMeta(**attrs)

    def with_changes(
        self,
        **kwargs: Unpack[CommandMetaProperties],
    ) -> "CommandMeta":
        attrs = merge_dicts(get_self_dict_restricted(self, CommandMeta), kwargs)
        return CommandMeta(**attrs)  # type: ignore[argument]


@dataclass
class Command:
    """
    Represents a command that can be triggered by a hotkey, with an attached handler.
    """

    info: CommandMeta
    handler: "FuncHotkeyHandler"
    no_intercept: bool

    def __get__(
        self, instance: object | None = None, owner: type | None = None
    ) -> "Command":
        return self

    def __str__(self):
        return f"{self.info.emoji} {self.info.label}"

    def bind(self, hotkey: "Hotkey"):
        from ..bindings import Binding

        return Binding(hotkey.info, self)

    def intercept(self, *interceptors: "FuncHotkeyInterceptor"):
        if self.no_intercept:
            return self
        handler = self.handler
        for interceptor in interceptors:
            handler = _wrap_interceptor(interceptor, handler)
        return Command(
            info=self.info, handler=handler, no_intercept=self.no_intercept
        )


class FuncHotkeyHandler(Protocol):
    """
    Represents a non-instance function that handles a hotkey event.
    """

    def __call__(self, event: "HotkeyEvent", /) -> Any: ...


class MethodHotkeyHandler(Protocol):
    """
    Represents an instance method that handles a hotkey event.
    """

    def __call__(self, instance: Any, event: "HotkeyEvent", /) -> Any: ...


type HotkeyHandler = FuncHotkeyHandler | MethodHotkeyHandler


class CommandProducer:
    """
    Represents an object that produces a command with an attached handler. When applied to a method,
    requires additional context to function properly.
    """

    def __str__(self):
        return str(self.cmd)

    def __init__(
        self,
        func: HotkeyHandler,
        cmd: CommandMeta,
        interceptor: "AnyHotkeyInterceptor | bool",
    ):
        self.func = func
        self.cmd = cmd
        self.interceptor = interceptor

    def _make(self, instance: object | None = None):
        def wrapper(event: "HotkeyEvent", /) -> Any:
            arg_count = len(inspect.signature(self.func).parameters)
            match arg_count:
                case 1:
                    return self.func(event)  # type: ignore
                case 2:
                    return self.func.__get__(instance)(event)  # type: ignore
                case _:
                    raise ValueError("Invalid number of arguments")

        existing_interceptor = self.interceptor
        cmd = self.cmd
        if callable(existing_interceptor) and is_unbound_method(
            existing_interceptor
        ):
            wrapper = _wrap_interceptor(
                partial(existing_interceptor, instance), wrapper
            )

        return Command(
            info=cmd, handler=wrapper, no_intercept=self.interceptor is not True
        )

    def __get__(
        self, instance: object | None, owner: type | None = None
    ) -> Command:
        """
        May bind an instance to a command handler so it can be invoked as an instance method.
        """
        return self._make(instance)


type CommandOrProducer = Command | CommandProducer


@dataclass(kw_only=True)
class command(CommandMeta):
    interceptor: "AnyHotkeyInterceptor | bool" = field(default=True)

    """
    Use this decorator on a method or function to turn it into a Command.
    """

    def __call__(self, handler: HotkeyHandler) -> "CommandProducer":
        cmd = self.with_defaults(label=clean_method_name(handler))
        return CommandProducer(handler, cmd=cmd, interceptor=self.interceptor)


def _wrap_interceptor(
    interceptor: "FuncHotkeyInterceptor", handler: FuncHotkeyHandler
) -> FuncHotkeyHandler:
    from keyweave.interception import HotkeyInterceptionEvent

    async def _handler(e: "HotkeyEvent"):
        interception = HotkeyInterceptionEvent(e, handler)
        result = interceptor(interception)
        if isinstance(result, Awaitable):
            await result
        if not interception.handled:
            raise ValueError(f"Interceptor {interceptor} did not handle {e}")
        return result

    return _handler



from typing import Any, Iterable

from typing import Callable

from pyee import EventEmitter


from keyweave.bindings import BindingCollection
from keyweave.bindings import Binding
from keyweave.commanding import Command, CommandProducer
from keyweave.hotkey import Hotkey
from keyweave.interception import (
    FuncHotkeyInterceptor,
    HotkeyInterceptor,
)
from keyweave.key_types import Key
from keyweave._hook import KeyHook
from keyweave.print.style import style
from keyweave.scheduling import default_scheduler, Scheduler
import sys


class Layout(EventEmitter):
    """
    A Hotkey layout consisting of `Binding` objects that link Hotkey events to Commands with attached handlers.

    Creating a layout using functions:

    >>> from pykeys import key, command
    ... @command("command_1")
    ... def command_1(e: HotkeyEvent):
    ...     pass
    ...
    ... @command("command_2")
    ... def command_2(e: HotkeyEvent):
    ...     pass
    >>> layout = Layout.create("my_layout", {
    ...     key.a: command_1,
    ...     key.b: command_2,
    ... })

    To enable a layout, use a context manager:
    >>> with layout:
    ...     import time
    ...     time.sleep(10)  # keep the layout active for 10 seconds

    You can also create a layout using a class by subclassing `LayoutClass`.
    """

    _registered: list[KeyHook]
    _map: BindingCollection
    _active: bool = False

    def __init__(
        self,
        *,
        name: str,
        scheduler: Scheduler | None = None,
        on_error: Callable[[BaseException], None] | None = None,
        bindings: Iterable[Binding] = (),
    ):
        super().__init__()

        def default_on_error(e: BaseException):
            import traceback

            # Print the full exception with traceback to stderr so the error is visible.
            traceback.print_exception(
                type(e), e, e.__traceback__, file=sys.stderr
            )

        scheduler = scheduler or default_scheduler(on_error or default_on_error)
        self.name = name
        self._scheduler = scheduler
        self._map = BindingCollection()
        for binding in bindings:
            self.add_binding(binding)

    def __iadd__(self, binding: Binding):
        self.add_binding(binding)
        return self

    @property
    def is_empty(self):
        return len(self._map) == 0

    def intercept(self, interceptor: FuncHotkeyInterceptor) -> "Layout":
        return Layout(
            name=self.name,
            scheduler=self._scheduler,
            bindings=[
                (
                    binding.intercept(interceptor)
                    if not binding.command.no_intercept
                    else binding
                )
                for binding in self._map.bindings
            ],
        )

    @property
    def active(self):
        return self._active

    def add_binding(self, binding: Binding):
        self._map += binding

    def __len__(self):
        return len(self._map)

    @property
    def bindings(self):
        return self._map.bindings

    def __iter__(self):
        return iter(self._map.bindings)

    def _get_key_hooks(self):
        return [
            KeyHook(key, bindings, self._scheduler)
            for key, bindings in self._map.pairs
        ]

    def __enter__(self):
        from ._print_layout import print_entering_message

        print_entering_message(self)
        key_hooks = self._get_key_hooks()
        registered: list[KeyHook] = []
        try:
            for hook in key_hooks:
                hook.__enter__()
                registered.append(hook)
            self.emit("enter", self)
        except:
            for hook in registered:
                hook.__exit__()
            raise
        self._registered = key_hooks

    def __exit__(self, *args: Any):
        for hook in self._registered:
            hook.__exit__()
        self.emit("exit", self)
        return False

    @staticmethod
    def create(
        name: str, d: dict[Hotkey | Key, Command | CommandProducer]
    ) -> "Layout":
        clean_dict = {k.__hotkey__().info: v for k, v in d.items()}
        xs = [Binding(k, v.__get__()) for k, v in clean_dict.items()]
        return Layout(name=name, bindings=xs)



from operator import is_
from typing import Any
from keyboard import is_pressed as kb_is_pressed


def _get_windows_id_for_mouse_key(key_id: str) -> int:
    return int(key_id.replace("mouse:", ""))


def _get_mouse_id_for_mouse_key(key_id: str) -> str:
    from mouse import LEFT, RIGHT, X, X2, MIDDLE  # type: ignore

    match key_id:
        case "mouse:1":
            return LEFT
        case "mouse:2":
            return RIGHT
        case "mouse:3":
            return MIDDLE
        case "mouse:4":
            return X
        case "mouse:5":
            return X2
        case _:
            raise ValueError("Key is not a recognized mouse key")


def _is_mouse_pressed_win32(key_id: str):
    from win32api import GetAsyncKeyState  # type: ignore

    id = _get_windows_id_for_mouse_key(key_id)
    pr: Any = GetAsyncKeyState(id) & 0x8000
    return pr


def _is_mouse_pressed_mouse(key_id: str):
    from mouse import is_pressed  # type: ignore

    mouse_id = _get_mouse_id_for_mouse_key(key_id)
    return is_pressed(mouse_id)


def _is_mouse_pressed(key_id: str):
    try:
        return _is_mouse_pressed_win32(key_id)
    except ImportError:
        return _is_mouse_pressed_mouse(key_id)


def _is_keyboard_pressed(key_id: str):
    hook_id = get_keyboard_hook_id(key_id)
    return kb_is_pressed(hook_id)


def is_key_pressed(key: str):
    if key.startswith("mouse:"):
        return _is_mouse_pressed(key)
    else:
        return _is_keyboard_pressed(key)


def get_keyboard_hook_id(key_id: str) -> str:
    match key_id:
        case "num:enter":
            return "enter"
        case "num:dot" | "num:.":
            return "."
        case "num:star" | "num:*" | "num:multiply":
            return "*"
        case "num:plus" | "num:+":
            return "+"
        case "num:minus" | "num:-":
            return "-"
        case "num:slash" | "num:/":
            return "/"
        case "mouse:1":
            return "left"
        case "mouse:2":
            return "right"
        case "mouse:3":
            return "middlemouse"
        case "mouse:4":
            return "x"
        case "mouse:5":
            return "x2"
        case _:
            return key_id.replace("num:", "num ")



from dataclasses import dataclass, field
import time
from typing import TYPE_CHECKING

from keyweave.commanding import CommandProducer
from keyweave.key_types import KeyInputState, KeySet, KeysInput

if TYPE_CHECKING:
    from keyweave.commanding import Command
    from keyweave.bindings import Binding


@dataclass(order=True, eq=True, frozen=True, unsafe_hash=True)
class HotkeyInfo:
    """
    Represents declarative information about a hotkey, including its trigger key, event type, and modifiers.
    """

    __match_args__ = ("trigger", "type", "modifiers", "passthrough")
    trigger: KeyInputState
    modifiers: KeySet = field(default=KeySet())
    passthrough: bool = field(default=False, compare=False)

    @property
    def specificity(self) -> int:
        """
        The specificity of a Hotkey. Used to decide which Hotkey to invoke when multiple are applicable.
        """
        return self.trigger.specificity + self.modifiers.specificity

    @property
    def trigger_label(self) -> str:
        """
        A label for the Hotkey's trigger key.
        """
        return str(self.trigger)

    def __str__(self) -> str:
        if not self.modifiers:
            return str(self.trigger)
        else:
            return f"{self.trigger_label} & {self.modifiers}"

    def __hotkey__(self):
        return Hotkey(self)


@dataclass(order=True, eq=True, frozen=True, unsafe_hash=True)
class Hotkey:
    """
    Represents a hotkey that can be transformed and bound to a command.
    """

    info: HotkeyInfo

    def __hotkey__(self):
        return Hotkey(self.info)

    def passthrough(self, enable: bool = True):
        """
        Whether the Hotkey should emit the Trigger key event.
        """
        return Hotkey(
            HotkeyInfo(
                trigger=self.info.trigger,
                modifiers=self.info.modifiers,
                passthrough=enable,
            )
        )

    @property
    def is_down(self) -> bool:
        """
        Whether the hotkey is triggered by a key press event.
        """
        return self.info.trigger.is_down

    def __call__(self, cmd: "Command | CommandProducer"):
        """
        Binds a command or command producer to the hotkey. Typically used as a decorator.
        """
        from ..bindings import BindingProducer

        return BindingProducer(cmd, self)

    @property
    def is_up(self) -> bool:
        """
        Whether the hotkey is triggered by a key release event.
        """
        return self.info.trigger.is_up

    def __and__(self, other: KeysInput):
        """
        Adds modifiers to a Hotkey.
        """
        return self.modifiers(other)

    def modifiers(self, modifiers: KeysInput):
        """
        Adds modifiers to a Hotkey.
        """
        return Hotkey(
            HotkeyInfo(
                trigger=self.info.trigger,
                modifiers=self.info.modifiers + modifiers,
                passthrough=self.info.passthrough,
            )
        )


type HotkeyInput = Hotkey | HotkeyInfo


@dataclass
class InputEvent:
    """
    Represents a low-level input event, such as a key press or release.
    """

    timestamp: float | None = None

    def __post_init__(self):
        if self.timestamp is None:
            self.timestamp = time.time()


@dataclass
class HotkeyEvent:
    """
    Represents a high-level event triggered by a hotkey, with a matching Command.
    """

    binding: "Binding"
    event: InputEvent

    @property
    def hotkey(self) -> "HotkeyInfo":
        return self.binding.hotkey

    @property
    def command(self) -> "Command":
        return self.binding.command

    def __str__(self):
        dt = time.strftime("%H:%M:%S", time.localtime(self.event.timestamp))
        return f"<HotkeyEvent {dt} :: {self.binding}>"



from dataclasses import dataclass, field
import inspect
from typing import Any, Awaitable, Iterable, Iterator, overload

from keyweave.commanding import (
    Command,
    CommandProducer,
    FuncHotkeyHandler,
)
from keyweave.interception import FuncHotkeyInterceptor, HotkeyInterceptionEvent
from keyweave.key_types import Key, KeyInputState
from keyweave.hotkey import (
    Hotkey,
    HotkeyEvent,
    HotkeyInfo,
    HotkeyInput,
    InputEvent,
)


@dataclass(match_args=True)
class Binding:
    hotkey: HotkeyInfo
    command: Command
    _number_of_args: int = field(init=False)

    def __str__(self):
        return f"⛓️‍💥{self.hotkey} → ⚡{self.command.info.label}"

    @property
    def handler(self) -> FuncHotkeyHandler:
        return self.command.handler

    def __post_init__(self):
        self._number_of_args = len(inspect.signature(self.handler).parameters)
        if not callable(self.handler):
            raise ValueError(f"handler must be a callable, got {self.handler}")
        if self._number_of_args != 1:
            raise ValueError(
                f"handler must accept 1 argument, got {self._number_of_args}"
            )

    def __call__(self, event: InputEvent, /):
        handler = self.handler
        triggered_key_event = HotkeyEvent(self, event)
        return handler(triggered_key_event)

    def intercept(self, *interceptors: FuncHotkeyInterceptor):
        handler = self.handler
        return Binding(
            self.hotkey,
            Command(
                info=self.command.info,
                handler=handler,
                no_intercept=self.command.no_intercept,
            ).intercept(*interceptors),
        )


class BindingCollection(Iterable["KeyBindingCollection"]):
    _map: dict[Key, "KeyBindingCollection"]
    _handler_map: dict[KeyInputState, Any]

    def __init__(self, input: dict[Key, "KeyBindingCollection"] = {}):
        self._map = input

    def __add__(self, input: Binding):
        trigger_key = input.hotkey.trigger.key
        new_map = self._map.copy()
        trigger_collection = new_map.get(
            trigger_key, KeyBindingCollection(trigger_key)
        )
        trigger_collection = trigger_collection.set(input)
        new_map[trigger_key] = trigger_collection
        return BindingCollection(new_map)

    @property
    def keys(self) -> set[Key]:
        return set(self._map.keys())

    @property
    def bindings(self) -> Iterable[Binding]:
        return (binding for collection in self for binding in collection)

    @property
    def pairs(self) -> Iterable[tuple[Key, "KeyBindingCollection"]]:
        return self._map.items()

    @overload
    def __getitem__(self, key: Key) -> "KeyBindingCollection": ...

    @overload
    def __getitem__(self, key: Hotkey) -> Binding: ...

    def __getitem__(
        self, key: Key | HotkeyInfo | Hotkey | KeyInputState
    ) -> "KeyBindingCollection | Binding":
        match key:
            case HotkeyInfo():
                return self._map[key.trigger.key][key]
            case Key() | KeyInputState():
                return self._map[key.__keystate__().key]
            case Hotkey():
                return self._map[key.info.trigger.key][key]

    def __iter__(self) -> Iterator["KeyBindingCollection"]:
        return iter(self._map.values())

    def __len__(self) -> int:
        return len(self._map)

    def __repr__(self) -> str:
        return repr(self._map)


class KeyBindingCollection:
    key: Key
    _map: dict[HotkeyInfo, Binding]

    def __init__(self, trigger: Key, bindings: dict[HotkeyInfo, Binding] = {}):
        self.key = trigger
        self._map = bindings

    def __getitem__(self, key: HotkeyInput) -> Binding:

        return self._map[key.__hotkey__().info]

    def set(self, binding: Binding):
        return KeyBindingCollection(
            self.key, {**self._map, binding.hotkey: binding}
        )

    def __len__(self) -> int:
        return len(self._map)

    def __iter__(self) -> Iterator[Binding]:
        return iter(self._map.values())

    def __add__(self, binding: Binding):
        return self.set(binding)


@dataclass
class BindingProducer:
    cmd: "Command | CommandProducer"
    hotkey: Hotkey

    def __get__(self, instance: object, owner: type) -> "Binding":
        r_cmd = self.cmd.__get__(instance, owner)

        return r_cmd.bind(self.hotkey)



from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, Iterable, Literal, overload

from termcolor import colored

from keyweave.print.escape import UnicodeEncoder

type _ColorName = Literal[
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "light_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "dark_grey",
    "black",
]

type _AnsiStyle = Literal[
    "bold",
    "dark",
    "underline",
    "blink",
    "reverse",
    "concealed",
    "strike",
]

type _Unset = Literal["unset"]
type _CharStyle = Literal[
    "uppercase",
    "lowercase",
]


def _alter_value[X](my_value: X | None, input: X | _Unset | None) -> X | None:
    if input == "unset":
        return None
    if input is not None:
        return input
    return my_value


@dataclass
class style:
    color: _ColorName | None = field(default=None)
    on_color: _ColorName | None = field(default=None)
    styles: Iterable[_AnsiStyle] | None = field(default=None)
    char_style: _CharStyle | None = field(default=None)
    padding: int = field(default=0)

    @property
    def _real_on_color(self) -> str | None:
        return f"on_{self.on_color}" if self.on_color else None

    def _apply_char_transform(self, text: str) -> str:
        if self.padding > 0:
            text = " " * self.padding + text + " " * self.padding
        match self.char_style:
            case "uppercase":
                text = text.upper()
            case "lowercase":
                text = text.lower()
            case None:
                pass
            case str:  # type: ignore[unreachable]
                raise NotImplementedError(str)

        return text

    def altered(
        self,
        *,
        color: _ColorName | None | _Unset = None,
        on_color: _ColorName | None | _Unset = None,
        styles: Iterable[_AnsiStyle] | None = None,
        char_style: _CharStyle | None | _Unset = None,
        padding: int | None = None,
    ) -> "style":
        return style(
            color=_alter_value(self.color, color),
            on_color=_alter_value(self.on_color, on_color),
            styles=_alter_value(self.styles, styles),
            char_style=_alter_value(self.char_style, char_style),
            padding=padding if padding is not None else self.padding,
        )

    def _apply_to_string(self, text: str) -> str:
        if text == "":
            return ""
        x = colored(
            text=self._apply_char_transform(text),
            color=self.color,
            on_color=self._real_on_color,
            attrs=list(self.styles) if self.styles else None,
        )

        return x

    @overload
    def __call__(self, text: str, /) -> str: ...

    @overload
    def __call__[**P](
        self, text_function: Callable[P, str], /
    ) -> Callable[P, str]: ...

    def __call__(
        self, text_or_function: str | Callable[..., str], /
    ) -> str | Callable[..., str]:

        if isinstance(text_or_function, str):
            return self._apply_to_string(text_or_function)
        elif callable(text_or_function):

            def wrapper(*args: Any, **kwargs: Any) -> str:
                return self._apply_to_string(text_or_function(*args, **kwargs))

            return wrapper
        else:
            raise ValueError(
                f"Expected a string or a callable, got {type(text_or_function)}"
            )



from typing import Awaitable, Callable
from keyweave.interception import HotkeyInterceptionEvent

from abc import ABC

from keyweave.bindings import Binding, BindingProducer
from keyweave.commanding import Command, CommandMeta, CommandProducer
from keyweave.layout._layout import Layout
from keyweave.scheduling import Scheduler
from keyweave._util.logging import keyweaveLogger
from keyweave.shorthand import SimpleCoroutine
from keyweave.util.reflect import get_attrs_down_to


my_logger = keyweaveLogger


class LayoutClass(ABC):
    """
    Use this to create a hotkey layout using a class. Using this mode, you declare commands and bind hotkeys.

    Commands are declared using the `command` decorator, while hotkeys are bound by using Hotkey objects as decorators.

    >>> from pykeys.layout import LayoutClass
    >>> from pykeys import key
    >>> from pykeys.commanding import command
    >>> class MyLayout(LayoutClass):
    ...     @(key.a & key.ctrl + key.shift)
    ...     @command(label="my_command")
    ...     def my_command(self, e: HotkeyEvent):
    ...         pass

    When you create an instance of `MyLayout`, you will get a `Layout` object rather than the class you defined.

    >>> layout = MyLayout()
    >>> isinstance(layout, Layout)
    True

    However, an instance of your class is still created and will be passed to the `self` parameter of command methods.

    Note that you must apply both `command` and hotkey decorators. Using just one of them is an error.

    However, you can still have helper methods that are not hotkey commands. Just don't use either decorator. These methods
    can only be accessed internally within the LayoutClass, such as by hotkey commands.

    >>> class MyLayout(LayoutClass):
    ...     def my_helper_method(self):
    ...         pass
    ...     @(key.a & key.ctrl + key.shift)
    ...     @command(label="my_command")
    ...     def my_command(self, e: HotkeyEvent):
    ...         self.my_helper_method()

    You can define an `__intercept__` method in your layout class. This lets you capture hotkey events and modify their behavior. This can be used for logging and similar:

    >>> class MyLayout(LayoutClass):
    ...     def __intercept__(self, intercepted: HotkeyInterceptionEvent):
    ...         print(intercepted)

    """

    _layout: "Layout"

    def __post_init__(self):
        """
        Executed internally after an instance of this class is created, before a Layout is returned.
        """
        pass

    def __intercept__(
        self, intercepted: HotkeyInterceptionEvent
    ) -> None | SimpleCoroutine[None]:
        """
        The default interceptor. If overriden, intercepts all hotkey events for this layout that haven't
        configured interception.
        """
        pass

    def __new__(
        cls,
        name: str | None = None,
        scheduler: Scheduler | None = None,
        on_error: Callable[[BaseException], None] | None = None,
    ):
        """
        Always returns a Layout instance wrapping this class instance.
        """
        obj = super().__new__(cls)
        my_logger.info(f"Creating instance")
        layout = Layout(
            name=name or cls.__name__, scheduler=scheduler, on_error=on_error
        )
        obj._layout = layout
        obj.__post_init__()

        attrs = get_attrs_down_to(
            obj,
            LayoutClass,
            resolve_descriptor=lambda x: isinstance(
                x, (CommandProducer, Command, Binding, BindingProducer)
            ),
        )

        for _, value in attrs.items():
            match value:
                case Binding() as b:
                    layout.add_binding(b)
                case CommandMeta() as c:
                    my_logger.warning(
                        f"Command {c} is missing a hotkey binding decorator. It will be ignored."
                    )
                case _:
                    pass

        if hasattr(obj, "__intercept__"):
            layout = layout.intercept(obj.__intercept__)  # type: ignore
        return layout



from functools import total_ordering
from typing import TYPE_CHECKING, Iterable, Literal

if TYPE_CHECKING:
    from keyweave.commanding import Command, CommandProducer


@total_ordering
class Key:
    """
    Represents a single keyboard, mouse, or controller key input. You can access a preset list of
    these events via the `pykeys.key` import, for example:

    >>> from pykeys import key
    >>> key.a # represents the "a" key

    Otherwise, you can use this class to reference keys using a string:

    >>> Key("a") # also represents the "a" key

    - For mouse keys, use `Key("mouse:1")` through `Key("mouse:5")` for buttons 1-5.
    - For numpad keys, use `Key("num:0")` through `Key("num:9")` for keys 0-9.

    For symbol keys, use:

    >>> Key("+") # the + key
    >>> Key("num:+") # the + key on the numpad

    Key objects are comparable, hashable, and equatable.
    """

    id: str

    def __init__(self, input: "str | Key"):
        self.id = input if isinstance(input, str) else input.id

    @property
    def label(self):
        return self.id

    @property
    def is_numpad(self):
        return self.id.startswith("num:")

    def __hash__(self) -> int:
        return hash(self.id)

    def __keystate__(self):
        return self.down

    def __hotkey__(self):
        return self.down.__hotkey__()

    def __invert__(self):
        return KeyInputState(self, "up")

    @property
    def is_mouse(self):
        """
        Whether this key is a mouse key.
        """
        return self.id.startswith("mouse:")

    def __call__(self, cmd: "Command | CommandProducer"):
        """
        Binds a command or command producer to the hotkey. Typically used as a decorator.
        """
        from ..bindings import BindingProducer

        return BindingProducer(cmd, self.down.modifiers([]))

    @property
    def is_keyboard(self):
        """
        Whether this key is a keyboard key.
        """
        return not self.is_mouse

    def __lt__(self, other: "Key") -> bool:
        return self.id < other.id

    @property
    def specificity(self):
        return 1

    def __and__(self, other: "KeysInput"):
        """
        Creates a Hotkey using the left key as a trigger (down event) and the right keys as modifiers.

        >>> from pykeys import key
        >>> hotkey = key.a & key.ctrl + key.shift
        >>> hotkey2 = key.a & [key.ctrl, key.shift]
        >>> hotkey3 = key.a & key.ctrl
        """
        return self.down.modifiers(other)

    @property
    def down(self) -> "KeyInputState":

        return KeyInputState(self, "down")

    @property
    def up(self):
        return KeyInputState(self, "up")

    def __repr__(self) -> str:
        return f"{self.id}"

    def __str__(self) -> str:
        return repr(self)


type KeyStateName = Literal["down", "up"]


@total_ordering
class KeyInputState:
    """
    Represents key up or key down states.
    """

    __match_args__ = ("key", "state")

    @property
    def hotkey(self):
        return self.__hotkey__()

    def __str__(self) -> str:
        """
        A label for the Hotkey's trigger key.
        """
        return f"{self.state_char}{self.key}"

    def __repr__(self) -> str:
        return str(self)

    def __keystate__(self):
        return self

    def __hotkey__(self):
        from ..hotkey import Hotkey, HotkeyInfo

        return Hotkey(
            HotkeyInfo(trigger=self, modifiers=KeySet(), passthrough=False)
        )

    def __invert__(self):
        return KeyInputState(self.key, "up" if self.is_down else "down")

    def __init__(self, key: Key, state: KeyStateName):
        self.key = key
        self.state = state

    def __hash__(self) -> int:
        return hash(self.key) ^ hash(self.state)

    def modifiers(self, modifiers: "KeysInput"):
        from ..hotkey import Hotkey, HotkeyInfo

        """
        Adds modifiers to a Hotkey.
        """
        return Hotkey(
            HotkeyInfo(
                trigger=self,
                modifiers=KeySet(modifiers),
                passthrough=False,
            )
        )

    def __and__(self, other: "KeysInput"):
        return self.modifiers(other)

    @property
    def state_char(self) -> str:
        return "↓" if self.state == "down" else "↑"

    def __lt__(self, other: "KeyInputState") -> bool:
        return self.key < other.key and self.state < other.state

    @property
    def specificity(self) -> int:
        return self.key.specificity

    @property
    def is_down(self) -> bool:
        return self.state == "down"

    @property
    def is_up(self) -> bool:
        return self.state == "up"


type KeyInput = "str | Key"

type KeysInput = KeySet | Iterable[Key | KeyInputState]


@total_ordering
class KeySet:
    """
    An unordered collection of `KeyInputState` objects, used as a set of modifiers.
    """

    __match_args__ = ("set",)
    set: dict[Key, KeyInputState]

    def __invert__(self):
        return KeySet(key.__invert__() for key in self.set)

    def __init__(self, input: KeysInput = {}):
        match input:
            case KeySet():
                self.set = input.set
            case _:
                self.set = {
                    key.__keystate__().key: key.__keystate__() for key in input
                }

    def __hash__(self) -> int:
        return hash((x, y) for x, y in self.set.items())

    def __add__(
        self,
        other: "KeyInputState | Key | Iterable[Key | KeyInputState] | KeySet",
    ):
        """
        Combines two keys into a `KeySet`, an unordered collection of keys.
        """
        from keyweave.key_types import KeySet

        match other:
            case KeyInputState():
                return KeySet(self.set | {other.key: other})
            case KeySet():
                return KeySet(self.set | other.set)
            case _:
                raise TypeError(f"Invalid key input: {other}")

    def __bool__(self) -> bool:
        return bool(self.set)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, KeySet) and self.set == other.set

    def __lt__(self, other: "KeySet") -> bool:
        if self.set.keys() != other.set.keys():
            return self.set.keys() < other.set.keys()
        ordered_keys = sorted(self.set.keys())
        for key in ordered_keys:
            if self.set[key] != other.set[key]:
                return self.set[key] < other.set[key]
        return False

    @property
    def specificity(self) -> int:
        return sum(key.specificity for key in self.set)

    def __iter__(self):
        return iter(self.set.values())

    def __contains__(self, key: Key | KeyInputState) -> bool:
        return self.set[key.__keystate__().key] == key.__keystate__()

    def __len__(self) -> int:
        return len(self.set)

    def __repr__(self) -> str:
        if not self.set:
            return ""
        joined = " + ".join(repr(key) for key in self.set)
        return joined

    def __str__(self) -> str:
        if not self.set:
            return "[]"
        joined = ", ".join(str(key) for key in self.set.values())
        return f"{{{joined}}}"



# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingTypeStubs=false
# pyright: reportUnknownArgumentType=false
from collections.abc import Iterable
from itertools import groupby
import re
from typing import Any
from beautifultable import BeautifulTable, ALIGN_LEFT
from yaml import Dumper, dump

from keyweave.bindings import Binding
from keyweave.commanding import CommandMeta
from keyweave.key_types import KeyInputState, KeySet
from keyweave.layout._layout import Layout
from termcolor import colored

from textwrap import indent

from keyweave.print.style import style
from keyweave.print.yaml import render_yaml


def _build_dict_from_bindings(bindings: Iterable[Binding]) -> dict[str, Any]:
    result = {}

    @style(color="light_green", styles=["underline", "bold"])
    def trigger_block(key_state: KeyInputState):
        return f"{str(key_state)} &"

    @style(color="light_cyan", styles=["bold"])
    def modifiers_block_key(modifiers: KeySet):
        return f"{' + '.join(str(m) for m in modifiers)}" if modifiers else ""

    @style(color="light_green")
    def command_short(command: CommandMeta):
        return str(command.label)

    @style(color="light_magenta")
    def command_desc(command: CommandMeta):
        parts = []
        if command.emoji:
            parts.append(command.emoji + " ")
        parts.append(command.description or "—")
        return "".join(parts)

    def command_and_description(command: CommandMeta):
        return {command_short(command): f"_NL_{command_desc(command)}"}

    by_trigger = groupby(bindings, lambda b: b.hotkey.trigger)
    result = {}
    for trigger, group in by_trigger:
        trigger_dict = {}
        for binding in group:
            mod_key = modifiers_block_key(binding.hotkey.modifiers)
            cmd_desc = command_and_description(binding.command.info)
            if mod_key:
                trigger_dict[mod_key] = cmd_desc
            else:
                trigger_dict.update(cmd_desc)
        result[trigger_block(trigger)] = trigger_dict
    return result


def print_layout(layout: Layout) -> str:
    layout_dict = _build_dict_from_bindings(layout)
    return render_yaml(layout_dict)


def print_entering_message(layout: Layout) -> None:

    base_style = style(
        color="white",
        styles=[],
        char_style="uppercase",
        on_color="light_green",
    )

    @base_style
    def entering_layout():
        return "➡️  Entering layout: "

    @base_style.altered(char_style="unset", on_color="light_blue")
    def layout_name():
        return f"{layout.name}"

    def get_header():
        return f"{entering_layout()}{layout_name()}"

    heading = get_header()
    layout_str = print_layout(layout)
    layout_str_indented = indent(layout_str, "  ")
    print(heading)
    print(layout_str_indented)



import os
import re
from typing import Any, override
from yaml import Dumper, SafeDumper, dump, safe_dump
from yaml.nodes import ScalarNode as ScalarNode

from keyweave.print.escape import UnicodeEncoder

encoder = UnicodeEncoder()


type NestedStrDict = dict[str, str | NestedStrDict]


class BlockStrDumper(Dumper):
    @override
    def represent_scalar(
        self, tag: str, value: Any, style: str | None = None
    ) -> ScalarNode:
        if isinstance(value, str) and value.startswith("_NL_"):
            value = value.replace("_NL_", "", 1)
            if len(value) > 35 and not value.endswith("\n"):
                style = ">"
                value += "\n"
        return super().represent_scalar(tag, value, style)  # type: ignore[return-value]


def _print_pure_yaml(target: NestedStrDict) -> str:
    return dump(
        target,
        width=45,
        sort_keys=False,
        indent=2,
        allow_unicode=True,
        Dumper=BlockStrDumper,
    )


def _escape_dict_values(d: NestedStrDict) -> NestedStrDict:
    result: NestedStrDict = {}
    for k, v in d.items():
        k = encoder.to_(k)
        if isinstance(v, dict):
            result[k] = _escape_dict_values(v)
        else:
            result[k] = encoder.to_(v)
    return result


def _unescape_yaml_text(s: str) -> str:
    return encoder.from_(s)


def render_yaml(target: NestedStrDict) -> str:
    target = _escape_dict_values(target)
    printed = _print_pure_yaml(target)
    printed = _unescape_yaml_text(printed)
    return printed



from typing import Mapping


def default_dicts[K, V](
    base: Mapping[K, V], *defaults: Mapping[K, V]
) -> dict[K, V]:
    """
    Create a new dictionary by merging a base dictionary with default values.

    Args:
        base (Mapping[K, V]): The base dictionary whose values take precedence.
        defaults (Mapping[K, V]): The default values to use if not present in the base.

    Returns:
        dict[K, V]: A new dictionary containing the merged values.
    """
    result = dict[K, V]()
    for d in reversed(defaults):
        for k, v in d.items():
            if v is not None:
                result[k] = v
    for k, v in base.items():
        if v is not None:
            result[k] = v
    return result


def merge_dicts[K, V](*dicts: Mapping[K, V]) -> dict[K, V]:
    """
    Merge multiple dictionaries into a single dictionary. Later dictionaries take precedence over earlier ones.

    Args:
        *dicts (Mapping[K, V]): The dictionaries to merge.

    Returns:
        dict[K, V]: A new dictionary containing the merged values.
    """
    result = dict[K, V]()
    for d in dicts:
        for k, v in d.items():
            if v is not None:
                result[k] = v
    return result



import re

from keyweave.print.based_int import BasedInt

marker = "+-"


def get_pattern():
    escaped_pattern = re.escape(marker)
    pattern = rf"{escaped_pattern}(.*?){escaped_pattern}"
    return pattern


_percent_pat = re.compile(get_pattern())


class UnicodeEncoder:
    _char_to_code = dict[str, str]()
    _code_to_char = dict[str, str]()
    _last_id = BasedInt()

    def _register_char(self, ch: str) -> str:
        if ch not in self._char_to_code:
            self._last_id += 1
            self._char_to_code[ch] = str(self._last_id)
            self._code_to_char[self._char_to_code[ch]] = ch
        return self._char_to_code[ch]

    def get_len_of_decoded_string(self, s: str) -> int:
        return len(_percent_pat.sub("", s)) + len(_percent_pat.findall(s))

    def _should_encode(self, ch: str) -> bool:
        cp = ord(ch)
        return not (0x20 <= cp < 0x7F) or cp in (0x1B,)  # encode non-ASCII

    def _decode(self, code: str) -> str | None:
        return self._code_to_char.get(code, None)

    def to_(self, s: str) -> str:
        out = list[str]()
        for ch in s:
            if not self._should_encode(ch):  # keep ASCII as-is
                out.append(ch)
            else:
                out.append(self._register_char(ch).join((marker, marker)))
        return "".join(out)

    def from_(self, s: str) -> str:
        def repl(m: re.Match[str]) -> str:
            deciphered = self._decode(m.group(1))
            if deciphered is None:
                assert False, "Internal error: unknown code"
            return deciphered

        return _percent_pat.sub(repl, s)



import inspect
from typing import Any, Callable


def get_attrs_down_to(
    target: object,
    base: type,
    *,
    resolve_descriptor: Callable[[Any], bool] = lambda _: True
):
    attrs = dict[str, object]()
    for cls in target.__class__.mro():
        if cls is base:
            break
        for key, value in cls.__dict__.items():
            if key not in attrs:
                attrs[key] = (
                    getattr(target, key) if resolve_descriptor(value) else value
                )

    return attrs


def clean_method_name(func: Callable[..., Any]) -> str:
    if hasattr(func, "__func__"):
        func = func.__func__  # type: ignore
    return func.__name__.lstrip("_")


def is_unbound_method(func: Callable[..., Any]) -> bool:
    return (
        inspect.isfunction(func)
        and "." in func.__qualname__
        and "self" in inspect.signature(func).parameters
    )


def get_self_dict_restricted(target: object, base: type) -> dict[str, Any]:
    """
    Get the __dict__ of an instance, but only including attributes defined in the instance's class and its subclasses down to (but not including) the specified base class.

    Args:
        target (object): The instance whose __dict__ is to be retrieved.
        base (type): The base class to stop at (not inclusive).

    Returns:
        dict[str, Any]: A dictionary containing the attributes of the instance.
    """
    return {
        k: v for k, v in target.__dict__.items() if k in base.__annotations__
    }



# pyright: reportUnusedImport=false
from keyweave.key_types import (
    Key,
    KeyInput,
    KeySet,
    KeysInput,
    KeyInputState,
)

from keyweave.layout._layout_class import LayoutClass
from keyweave.layout._layout import Layout
from keyweave import scheduling
from keyweave import key
from keyweave.hotkey import HotkeyEvent
from keyweave.commanding import command, Command, CommandProducer
from keyweave.key_types import Key
from keyweave.interception import HotkeyInterceptionEvent



# Add missing imports and logger
import asyncio
from concurrent.futures import ThreadPoolExecutor
import threading
import logging
from asyncio import AbstractEventLoop
from typing import Any, Callable, Coroutine

from keyweave.shorthand import SimpleCoroutine

logger = logging.getLogger("keyweave")


def create_event_loop(trace_name: str, workers: int = 1) -> AbstractEventLoop:

    loop = asyncio.new_event_loop()
    ready = threading.Event()
    pool = ThreadPoolExecutor(
        max_workers=workers, thread_name_prefix=f"{trace_name}"
    )

    def _run_loop() -> None:
        try:
            # Bind the loop to this thread
            asyncio.set_event_loop(loop)
            loop.set_default_executor(pool)
            # Set the executor while the loop is bound to the current thread
            # Signal that the loop/thread is ready
            ready.set()
            # Run the loop forever on this thread
            loop.run_forever()
        except Exception:
            logger.exception("Async loop crashed")

    t = threading.Thread(
        target=_run_loop, name=f"{trace_name}-loop", daemon=True
    )
    t.start()

    # Wait until the loop thread has bound the loop and set the executor
    ready.wait()
    return loop


def norm_maybe_async[**P, R](
    func: Callable[P, R | SimpleCoroutine[R]], /
) -> Callable[P, SimpleCoroutine[R]]:

    async def _wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
        result = func(*args, **kwargs)
        if isinstance(result, Coroutine):
            return await result  # type: ignore
        return result

    return _wrapped



from collections.abc import Coroutine
from dataclasses import dataclass
from typing import Any, Callable, Iterable, Literal, overload
from termcolor import COLORS, colored

type SimpleCoroutine[R] = Coroutine[Any, Any, R]



# pyright: reportMissingTypeStubs=false
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownMemberType=false
from baseconv import BaseConverter

converter = BaseConverter(
    "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
)


class BasedInt(int):
    """Integer-like object represented in the project's custom base.

    Construction:
    - BasedInt(int_or_str)
      If given a str, it is decoded using the module-level `converter`.
      If given an int (or something int()-able), it is used as-is.

    Implementation note:
    This subclasses ``int`` so it automatically supports all standard
    integer operations. ``__str__``/``__repr__`` are overridden to emit
    the base-converted representation.
    """

    def __new__(cls, value: 'int | str | "BasedInt"' = 0) -> "BasedInt":
        # If already a BasedInt, return it unchanged
        if isinstance(value, BasedInt):
            return value

        # Parse strings via the converter (handle optional leading '-')
        if isinstance(value, str):
            s = value
            if s.startswith("-"):
                val = -int(converter.decode(s[1:]))
            else:
                val = int(converter.decode(s))
        else:
            # coerce other numeric types to int
            val = int(value)

        return int.__new__(cls, val)

    def __str__(self) -> str:  # emit converted representation
        v = int(self)
        if v < 0:
            return "-" + converter.encode(-v)
        return converter.encode(v)

    def __repr__(self) -> str:  # same as __str__, per requirement
        return self.__str__()



from asyncio import AbstractEventLoop, Future
from typing import Any, Coroutine, Protocol
from dataclasses import dataclass, field

from keyweave.shorthand import SimpleCoroutine
from keyweave.util.event_loop import create_event_loop, norm_maybe_async


class ScheduleItem(Protocol):
    def __call__(self) -> None | SimpleCoroutine[None]: ...


class Scheduler(Protocol):

    def __call__(self, func: ScheduleItem, /) -> None: ...


class ScheduleErrorHandler(Protocol):
    def __call__(self, error: BaseException, /) -> None: ...


@dataclass(kw_only=True)
class AsyncIoScheduler(Scheduler):
    on_error: ScheduleErrorHandler
    workers: int
    _event_loop: AbstractEventLoop = field(init=False)

    def __post_init__(self):
        self._event_loop = create_event_loop(
            "KeyWeave-Scheduler", workers=self.workers
        )

    def __call__(self, func: ScheduleItem) -> None:
        def callback(future: Future[None]):
            ex = future.exception()
            if ex:
                self.on_error(ex)

        def run_item():
            self._event_loop.create_task(
                norm_maybe_async(func)()
            ).add_done_callback(callback)

        self._event_loop.call_soon_threadsafe(run_item)


def default_scheduler(on_error: ScheduleErrorHandler):
    return AsyncIoScheduler(workers=1, on_error=on_error)
